# -*- coding: utf-8 -*-
"""
Created on Wed Feb 12 18:48:25 2025

@author: sukin732
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.linalg import expm
from scipy.optimize import curve_fit

# ---- Loading the Data ----
File = "IrHH_data/kd_IrHH_DMSO_[3]/"
n = 'kd_IrHH_Ha_DMSO_100mM_[3]_norm.txt'

# Load the data from a .txt file (use delimiter='\t' or ' ' depending on your file format)
data_in = np.loadtxt(File + n, delimiter='\t')  # Adjust delimiter if needed

# Extract columns into separate variables
data = {
    'dmix0': data_in[:, 0],
    'dmix': data_in[:, 0] - data_in[0, 0],  # Adjust the first point to 0
    'Ha': data_in[:, 1],
    'Hb': data_in[:, 2],
    'H2': data_in[:, 3],
    'Ha2': data_in[:, 4],
    'Hb2': data_in[:, 5],
    'H3': data_in[:, 6]
}

# ---- The SABRE Hydrogen Exchange Function ----
def SABRE_hydrogen_exchange_Lmatrix_6(MH1a, MH1b, MH2a, MH2b, MH3ab, MH2, 
                                      kd1, kd2, kd3, kaprime1, kaprime2, kaprime3, 
                                      kex1, kex2, k12, k21, R1, R2, R3, Rf, x):
    # Rate matrix L
    L = np.array([[-kd1 - R1 - kex1 - k12, kex1, k21/2, k21/2, 0, kaprime1/2],
                  [kex1, -kd1 - R1 - kex1 - k12, k21/2, k21/2, 0, kaprime1/2],
                  [k12/2, k12/2, -kd2 - R2 - kex2 - k21, kex2, 0, kaprime2/2],
                  [k12/2, k12/2, kex2, -kd2 - R2 - kex2 - k21, 0, kaprime2/2],
                  [0, 0, 0, 0, -R3 - kd3, kaprime3],
                  [kd1, kd1, kd2, kd2, kd3, -Rf - kaprime1 - kaprime2 - kaprime3]])

    # Initial magnetization vector M0
    M0 = np.array([MH1a, MH1b, MH2a, MH2b, MH3ab, MH2])

    # Prepare empty lists to store the results
    y1 = np.zeros(len(x)//6)
    y2 = np.zeros_like(y1)
    y3 = np.zeros_like(y1)
    y4 = np.zeros_like(y1)
    y5 = np.zeros_like(y1)
    y6 = np.zeros_like(y1)

    # Loop over the data points and compute the results
    for ix in range(len(x)//6):
        M = np.dot(expm(L * x[ix]), M0)  # Matrix exponential multiplication
        y1[ix] = M[0]
        y2[ix] = M[1]
        y3[ix] = M[2]
        y4[ix] = M[3]
        y5[ix] = M[4]
        y6[ix] = M[5]

    # Combine the results into a single array
    y = np.concatenate([y1, y2, y3, y4, y5, y6])

    return y

# ---- Initial Guess for the Parameters ----
I = [0.4, 2.3, 1.29, 1, 1.53, 1, 2.3, 1.29, 0.05, 1.53, 1, 1, 1, 1]

# Initial values for the fitting parameters (adjusted from MATLAB)
MH1a = data['Ha'][0]
MH1b = data['Hb'][0]
MH2 = data['H2'][0]
MH2a = data['Ha2'][0]
MH2b = data['Hb2'][0]
MH3ab = data['H3'][0]

# ---- Prepare Data for Fitting ----
xdata_fit = np.repeat(data['dmix'][:, np.newaxis], 6, axis=1)  # Repeat data['dmix'] along columns (6 times)
xdata_fit = xdata_fit.flatten()  # Flatten the result to make it a 1D array
ydata = np.concatenate([data['Ha'], data['Hb'], data['H2'], data['Ha2'], data['Hb2'], data['H3']])

# ---- Fit the Model ----
def fitfunc(x, *params):
    # Unpack parameters
    MH1a, MH1b, MH2a, MH2b, MH3ab, MH2, kd1, kd2, kd3, kaprime1, kaprime2, kaprime3, kex1, kex2, k12, k21, R1, R2, R3, Rf = params
    
    # Call the SABRE function
    return SABRE_hydrogen_exchange_Lmatrix_6(MH1a, MH1b, MH2a, MH2b, MH3ab, MH2, 
                                             kd1, kd2, kd3, kaprime1, kaprime2, kaprime3, 
                                             kex1, kex2, k12, k21, R1, R2, R3, Rf, x)

# Fit the model using curve_fit
params, covariance = curve_fit(fitfunc, xdata_fit, ydata, p0=[MH1a, MH1b, MH2a, MH2b, MH3ab, MH2] + I)

# Print out the results
coefficients_standard_errors = np.sqrt(np.diag(covariance))

# Output results in a similar format as MATLAB
print("Fitting Results:")
print(f"kd1 = {params[6]:.3f} ± {coefficients_standard_errors[6]:.3f} s-1")
print(f"kd2 = {params[7]:.3f} ± {coefficients_standard_errors[7]:.3f} s-1")
print(f"kd3 = {params[8]:.3f} ± {coefficients_standard_errors[8]:.3f} s-1")
print(f"kaprime1 = {params[9]:.3f} ± {coefficients_standard_errors[9]:.3f} s-1")
print(f"kaprime2 = {params[10]:.3f} ± {coefficients_standard_errors[10]:.3f} s-1")
print(f"kaprime3 = {params[11]:.3f} ± {coefficients_standard_errors[11]:.3f} s-1")
print(f"kex1 = {params[12]:.3f} ± {coefficients_standard_errors[12]:.3f} s-1")
print(f"kex2 = {params[13]:.3f} ± {coefficients_standard_errors[13]:.3f} s-1")
print(f"k12 = {params[14]:.3f} ± {coefficients_standard_errors[14]:.3f} s-1")
print(f"k21 = {params[15]:.3f} ± {coefficients_standard_errors[15]:.3f} s-1")
print(f"R1 = {params[16]:.3f} ± {coefficients_standard_errors[16]:.3f} s-1")
print(f"R2 = {params[17]:.3f} ± {coefficients_standard_errors[17]:.3f} s-1")
print(f"R3 = {params[18]:.3f} ± {coefficients_standard_errors[18]:.3f} s-1")
print(f"Rf = {params[19]:.3f} ± {coefficients_standard_errors[19]:.3f} s-1")

# ---- Plot the Results ----
# Plot the experimental data and the fitting results

# Create a figure
plt.figure(figsize=(10, 8), dpi=80)

# Plot the experimental data
plt.semilogx(data['dmix0'], data['Ha'], 'ob', label='Ha (Experimental)', markersize=8)  # Blue circles for Ha
plt.semilogx(data['dmix0'], data['Hb'], 'or', label='Hb (Experimental)', markersize=8)  # Red circles for Hb
plt.semilogx(data['dmix0'], data['H2'], 'og', label='H2 (Experimental)', markersize=8)  # Green circles for H2
plt.semilogx(data['dmix0'], data['Ha2'], 'b--o', label='Ha2 (Experimental)', markersize=8)  # Blue dashed line with circle for Ha2
plt.semilogx(data['dmix0'], data['Hb2'], 'g--o', label='Hb2 (Experimental)', markersize=8)  # Green dashed line with circle for Hb2
plt.semilogx(data['dmix0'], data['H3'], 'r*', label='H3 (Experimental)', markersize=8)  # Red star marker for H3

# Plot the fitted data
plt.semilogx(xdata_fit, ydata[:len(xdata_fit)], 'k-', label='Fitted Model')  # Black line for fitted model

# Set axis limits
plt.ylim([-0.2, 1.1])  # Adjust the Y-axis limits as needed
plt.xlim([data['dmix0'][0], data['dmix0'][-1]])  # X-axis limits based on data

# Add labels and title
plt.xlabel('Dmix (s)', fontsize=16)
plt.ylabel('Intensity', fontsize=16)
plt.title('Experimental Data vs. Fitted Model', fontsize=18)

# Show the legend
plt.legend(fontsize=14)

# Show grid for better readability
plt.grid(True, which="both", linestyle='--', linewidth=0.5)

# Show the plot
plt.show()

# Optional: Save the figure if needed
plt.savefig("fit_results.png")
