% [1] Chiavazza et al 10.1016/j.jmr.2012.11.016
% [2] Kharkov et al 10.1039/d1cp05537b

clear all; close all; clc; % Start fresh

%% Parameter Configuration =============================================
paramList =      {'J', 'tauJ', 'tauD', 'tauCSA', 'CSA',     'rIS',  'T10'};

initialGuesses = [
    10,     % J (Hz) <-- Start with smaller J again
    5e-11,  % tauJ (s)
    5e-11,  % tauD (s) <-- Represents tauC
    5e-11,  % tauCSA (s) - Guess ignored if constrained
    160e-6, % CSA (unitless ~ppm*1e-6)
    2.2e-10,% rIS (m) <-- Initial guess for rIS
    6       % T10 (s)
];

% --- Define which parameters to FIX and their FIXED VALUES ---
constrainTauDtoTauCSA = false; % ADD FLAG TO CONTROL CONSTRAINT

fixedFlags = [
    false,  % J
    false,  % tauJ
    false,  % tauD  <-- FIT (Represents tauC)
    true,   % tauCSA <-- FIX (Will be set equal to tauD)
    false,  % CSA
    false,   % rIS   <-- Keep rIS Fixed
    false   % T10
];
fixedValues = initialGuesses; % Start with initial guesses
% --- Set the fixed values ---
% fixedValues(strcmp(paramList, 'tauD')) = 1e-30; % REMOVED - tauD is fitted
fixedValues(strcmp(paramList, 'rIS')) = 2.2e-10; % Fix rIS based on geometry guess
% Note: fixedValue for tauCSA doesn't strictly matter when constrainTauDtoTauCSA is true

% --- Physically Motivated Bounds for Pyridine ---
lowerBounds = [
    1,      % J
    1e-12,  % tauJ
    1e-12,  % tauD <-- Lower bound for tauC
    1e-12,  % tauCSA - Bound ignored if constrained
    50e-6,  % CSA (50ppm)
    1.8e-10,% rIS
    0.1     % T10
];
upperBounds = [
    1000,    % J
    1e-6,   % tauJ
    1e-8,   % tauD <-- Upper bound for tauC
    1e-7,   % tauCSA - Bound ignored if constrained
    400e-6, % CSA (400ppm)
    2.8e-10,% rIS
    50      % T10
];

% --- REMOVED SCALING FACTORS ---
% =====================================================================
%% Constants
S_init = 1/2;
gamma1H = 26.75222080e7;
gamma15N = -2.7126189e7;

%% Experimental data
D = [
0.032	0.35425
0.048	0.35168
0.064	0.52731
0.096	0.95071
0.128	1.57836
0.192	2.39362
0.256	3.43533
0.384	4.2081
0.512	5.61985
0.768	5.97045
1.024	5.65204
1.536	6.86752
2	6.69743
2.5	6.54266
3	6.63703
3.5	6.00461
4	6.1867
4.5	5.73373
5	5.09948
6	5.55681
7	4.60855
8	3.25459
9	3.60895
];
B_exp = D(:,1);
T1_exp = D(:,2);
R1_exp = 1./T1_exp;
B = logspace(log10(min(B_exp)*0.1), log10(max(B_exp)*1.1), 1000)';


%% Prepare for Fitting (NO SCALING)
fitIndices = find(~fixedFlags);
fixedIndices = find(fixedFlags);

params_base_unscaled = fixedValues;
% If tauCSA is constrained, make sure its value in the base vector is consistent
if constrainTauDtoTauCSA
    tauD_idx = find(strcmp(paramList, 'tauD')); % Use index finding for robustness
    tauCSA_idx = find(strcmp(paramList, 'tauCSA'));
    if ~isempty(tauD_idx) && ~isempty(tauCSA_idx)
        params_base_unscaled(tauCSA_idx) = params_base_unscaled(tauD_idx); % Ensure base tauCSA matches base tauD
    end
end
params_base_unscaled(fitIndices) = initialGuesses(fitIndices);

params0_fit_unscaled = params_base_unscaled(fitIndices);
lb_fit_unscaled = lowerBounds(fitIndices);
ub_fit_unscaled = upperBounds(fitIndices);

%% Perform the fitting (NO SCALING)
options = optimoptions('lsqnonlin', ...
    'Display', 'iter', ...
    'MaxFunctionEvaluations', 5000, ...
    'MaxIterations', 1000, ...
    'FunctionTolerance', 1e-9, ...
    'StepTolerance', 1e-9, ...
    'OptimalityTolerance', 1e-9, ...
    'Algorithm', 'trust-region-reflective', ...
    'FiniteDifferenceType', 'central', ...
    'UseParallel', false);

% Pass constrainTauDtoTauCSA flag
objFun_unscaled = @(p_unscaled_fit) calculate_R1_dynamic_unscaled(p_unscaled_fit, params_base_unscaled, fixedFlags, constrainTauDtoTauCSA, B_exp, gamma15N, gamma1H) - R1_exp;

[params_fit_unscaled, resnorm, residual, exitflag, output, ~, jacobian_unscaled] = lsqnonlin(objFun_unscaled, params0_fit_unscaled, lb_fit_unscaled, ub_fit_unscaled, options);

% Condition number check
if exist('jacobian_unscaled','var') && ~isempty(jacobian_unscaled)
    try
        jacobian_full = full(jacobian_unscaled);
        cond_num = cond(jacobian_full' * jacobian_full);
        fprintf('Condition number of J''*J: %.2e\n', cond_num);
    catch cond_err
        fprintf('Could not calculate condition number: %s\n', cond_err.message);
    end
else
    fprintf('Jacobian not available for condition number check.\n');
end


%% Process Results (NO SCALING)
% Reconstruct the full UNSCALED parameter vector
params_all_final = params_base_unscaled;
params_all_final(fitIndices) = params_fit_unscaled; % Insert fitted UNSCALED values

% *** ADDED: Re-apply constraint AFTER fit for consistent reporting ***
if constrainTauDtoTauCSA
    tauD_idx = find(strcmp(paramList, 'tauD'));
    tauCSA_idx = find(strcmp(paramList, 'tauCSA'));
     if ~isempty(tauD_idx) && ~isempty(tauCSA_idx)
        params_all_final(tauCSA_idx) = params_all_final(tauD_idx); % Ensure final tauCSA = final tauD
     end
end
% ********************************************************************

finalValues = struct();
for i = 1:length(paramList)
    finalValues.(paramList{i}) = params_all_final(i);
end

% Confidence intervals
param_errors_fit = nan(size(params_fit_unscaled));
ci_calculation_ok = false;
if exitflag > 0
    try
        ci_unscaled_fit = nlparci(params_fit_unscaled, residual, 'jacobian', jacobian_unscaled);
        param_errors_fit = (ci_unscaled_fit(:,2) - ci_unscaled_fit(:,1)) / 2;
        ci_calculation_ok = true;
    catch ME
        warning('Could not calculate confidence intervals using nlparci: %s', ME.message);
        if exist('jacobian_unscaled','var') && ~isempty(jacobian_unscaled)
             try
                 cond_num = cond(full(jacobian_unscaled)' * full(jacobian_unscaled));
                 fprintf('Condition number of J''*J: %.2e\n', cond_num);
             catch cond_err
                 fprintf('Could not calculate condition number: %s\n', cond_err.message);
             end
        end
    end
end

% Display results
fprintf('\n--- Fit Results ---\n');
R2 = 1 - resnorm/sum((R1_exp - mean(R1_exp)).^2);
fprintf('R-squared: %.4f\n', R2);
fprintf('Exit Flag: %d, Message: %s\n', exitflag, output.message);
if constrainTauDtoTauCSA; fprintf('Constraint Applied: tauD = tauCSA\n'); end % Corrected message
fprintf('\n--- Final Parameter Values and 95%% CIs ---\n');
err_idx = 1;
for i = 1:length(paramList)
    pName = paramList{i}; pValue = finalValues.(pName);
    % Special handling for constrained tauCSA display
    isTauCSAConstrained = constrainTauDtoTauCSA && strcmp(pName, 'tauCSA');

    if fixedFlags(i) && ~isTauCSAConstrained % Fixed parameters (excluding constrained tauCSA)
        fprintf('%s (fixed) = %.3e\n', pName, pValue);
    elseif isTauCSAConstrained % Constrained tauCSA
        fprintf('%s (constrained to tauD) = %.3e\n', pName, pValue); % Corrected display
    else % Fitted parameters
        if ci_calculation_ok && err_idx <= length(param_errors_fit)
            pError = param_errors_fit(err_idx);
            fprintf('%s = %.3e ± %.3e\n', pName, pValue, pError);
        else
            fprintf('%s = %.3e ± (CI failed)\n', pName, pValue);
        end
        err_idx = err_idx + 1;
    end
end

% Check bounds
fprintf('\nChecking if fitted parameters are close to bounds:\n');
tol = 1e-5; fit_idx_counter = 1;
for i = 1:length(paramList)
    % Only check parameters that were actually fitted by lsqnonlin
    if ~fixedFlags(i)
        param_name = paramList{i}; fitted_value = finalValues.(param_name);
        lb_val = lowerBounds(i); ub_val = upperBounds(i);
        if fitted_value <= lb_val * (1 + tol) + lb_val*tol
             fprintf('Parameter %s (%.3e) is close to lower bound (%.3e)\n', param_name, fitted_value, lb_val);
        elseif fitted_value >= ub_val * (1 - tol) - ub_val*tol
             fprintf('Parameter %s (%.3e) is close to upper bound (%.3e)\n', param_name, fitted_value, ub_val);
        end
        fit_idx_counter = fit_idx_counter + 1;
    end
end

% ... (Rest of script: Calculate Contributions, Plotting) ...
% ... (Helper function calculate_R1_dynamic_unscaled is correct) ...
%% Calculate relaxation rates & Contributions (Same as before)
% ... (Code is identical, uses finalValues struct) ...
R1_J_fit = R_J(finalValues.J, S_init, gamma15N, gamma1H, finalValues.tauJ, B);
R1_DD_fit = R_DD(finalValues.rIS, S_init, gamma15N, gamma1H, finalValues.tauD, B);
R1_CSA_fit = R_CSA(gamma15N*B, finalValues.CSA, finalValues.tauCSA);
R10_fit = 1/finalValues.T10;
R1_fit_total = R10_fit + R1_J_fit + R1_DD_fit + R1_CSA_fit;
R1_fit_at_exp = R10_fit + R_J(finalValues.J, S_init, gamma15N, gamma1H, finalValues.tauJ, B_exp) + R_DD(finalValues.rIS, S_init, gamma15N, gamma1H, finalValues.tauD, B_exp) + R_CSA(gamma15N*B_exp, finalValues.CSA, finalValues.tauCSA);
fprintf('\nRelative contributions at 1T:\n'); % ... (rest of contribution calculation) ...
R1_J_1T = R_J(finalValues.J, S_init, gamma15N, gamma1H, finalValues.tauJ, 1);
R1_DD_1T = R_DD(finalValues.rIS, S_init, gamma15N, gamma1H, finalValues.tauD, 1);
R1_CSA_1T = R_CSA(gamma15N*1, finalValues.CSA, finalValues.tauCSA);
R1_total_1T = R10_fit + R1_J_1T + R1_DD_1T + R1_CSA_1T;
fprintf('R10: %.4e s^-1 (%.1f%%)\n', R10_fit, 100*R10_fit/R1_total_1T);
fprintf('R1_J: %.4e s^-1 (%.1f%%)\n', R1_J_1T, 100*R1_J_1T/R1_total_1T);
fprintf('R1_DD: %.4e s^-1 (%.1f%%)\n', R1_DD_1T, 100*R1_DD_1T/R1_total_1T);
fprintf('R1_CSA: %.4e s^-1 (%.1f%%)\n', R1_CSA_1T, 100*R1_CSA_1T/R1_total_1T);
fprintf('R1_total: %.4e s^-1\n', R1_total_1T);
fprintf('\nRelative contributions at 9.4T:\n'); % ... (rest of contribution calculation) ...
R1_J_9T = R_J(finalValues.J, S_init, gamma15N, gamma1H, finalValues.tauJ, 9.4);
R1_DD_9T = R_DD(finalValues.rIS, S_init, gamma15N, gamma1H, finalValues.tauD, 9.4);
R1_CSA_9T = R_CSA(gamma15N*9.4, finalValues.CSA, finalValues.tauCSA);
R1_total_9T = R10_fit + R1_J_9T + R1_DD_9T + R1_CSA_9T;
fprintf('R10: %.4e s^-1 (%.1f%%)\n', R10_fit, 100*R10_fit/R1_total_9T);
fprintf('R1_J: %.4e s^-1 (%.1f%%)\n', R1_J_9T, 100*R1_J_9T/R1_total_9T);
fprintf('R1_DD: %.4e s^-1 (%.1f%%)\n', R1_DD_9T, 100*R1_DD_9T/R1_total_9T);
fprintf('R1_CSA: %.4e s^-1 (%.1f%%)\n', R1_CSA_9T, 100*R1_CSA_9T/R1_total_9T);
fprintf('R1_total: %.4e s^-1\n', R1_total_9T);

%% Plotting
close all
fixedParamStrParts = {};
constrainedParamStrParts = {};
for i = 1:length(paramList)
    isTauCSAConstrained = constrainTauDtoTauCSA && strcmp(paramList{i}, 'tauCSA');
    if fixedFlags(i) && ~isTauCSAConstrained
        fixedParamStrParts{end+1} = paramList{i};
    elseif isTauCSAConstrained
        constrainedParamStrParts{end+1} = 'tauD=tauCSA';
    end
end
fixedParamStr = strjoin(fixedParamStrParts, ', ');
constrainedParamStr = strjoin(constrainedParamStrParts, ', ');
plotTitleSuffix = '';
if ~isempty(fixedParamStr); plotTitleSuffix = [plotTitleSuffix, ', Fixed: ', fixedParamStr]; end
if ~isempty(constrainedParamStr); plotTitleSuffix = [plotTitleSuffix, ', Constrained: ', constrainedParamStr]; end


% R1 plot - all contributions
figure("Color", "White")
loglog(B, [R1_J_fit, R1_DD_fit, R1_CSA_fit, R1_fit_total], "LineWidth", 3);
xlabel("Magnetic field, T")
ylabel("R_1 (s^{-1})")
hold on
loglog(B_exp, R1_exp, "o", "MarkerSize", 10, "LineWidth", 2)
hold off
title(sprintf('R_1 (all contributions%s)', plotTitleSuffix))
legend(["R_{1J}", "R_{1DD}", "R_{1CSA}", "R_{1,all}", "Exp data"])
grid on; xlim([min(B_exp)*0.9, max(B_exp)*1.1]); % Auto Y limits usually fine

% T1 plot - all contributions
figure("Color", "White")
loglog(B, [1./R1_J_fit, 1./R1_DD_fit, 1./R1_CSA_fit, 1./R1_fit_total], "LineWidth", 3);
xlabel("Magnetic field, T")
ylabel("T_1 (s)")
hold on
loglog(B_exp, T1_exp, "o", "MarkerSize", 10, "LineWidth", 2)
hold off
title(sprintf('T_1 (all contributions%s)', plotTitleSuffix))
legend(["T_{1J}", "T_{1DD}", "T_{1CSA}", "T_{1,all}", "Exp data"])
grid on; xlim([min(B_exp)*0.9, max(B_exp)*1.1]);

% R1 plot - scaled to exp data
figure("Color", "White")
loglog(B, [R1_J_fit, R1_DD_fit, R1_CSA_fit, R1_fit_total], "LineWidth", 3);
xlabel("Magnetic field, T")
ylabel("R_1 (s^{-1})")
hold on
loglog(B_exp, R1_exp, "o", "MarkerSize", 10, "LineWidth", 2)
hold off
title(sprintf('R_1 (scaled to exp. data%s) - R² = %.4f', plotTitleSuffix, R2))
legend(["R_{1J}", "R_{1DD}", "R_{1CSA}", "R_{1,all}", "Exp data"])
grid on; xlim([min(B_exp)*0.9, max(B_exp)*1.1]); ylim([min(R1_exp)*0.9, max(R1_exp)*1.1]);

% Residuals plot
figure("Color", "White")
semilogx(B_exp, R1_exp - R1_fit_at_exp, 'o', 'MarkerSize', 10, 'LineWidth', 2);
xlabel("Magnetic field, T")
ylabel("Residuals (s^{-1})")
title(sprintf('Fit Residuals%s', plotTitleSuffix))
grid on
line([min(B_exp)*0.9, max(B_exp)*1.1], [0, 0], 'Color', 'k', 'LineStyle', '--')

% Confidence Interval Plots for FITTED parameters
if ci_calculation_ok
    try
        fittedParamNames = paramList(fitIndices);
        fittedParamValues = params_all_final(fitIndices);

        figure("Color", "White")
        errorbar(1:length(fittedParamValues), fittedParamValues, param_errors_fit, 'o', 'MarkerSize', 10, 'LineWidth', 2, 'LineStyle', 'none');
        set(gca, 'XTick', 1:length(fittedParamValues), 'XTickLabel', fittedParamNames);
        ylabel('Parameter Value');
        title(sprintf('Fitted Parameters CIs%s', plotTitleSuffix));
        grid on;
        set(gca,'YScale','log')

        figure("Color", "White")
        relative_uncertainty_fit = param_errors_fit ./ abs(fittedParamValues);
        bar(1:length(fittedParamValues), relative_uncertainty_fit);
        set(gca, 'XTick', 1:length(fittedParamValues), 'XTickLabel', fittedParamNames);
        ylabel('Relative Uncertainty (CI half-width / value)');
        title(sprintf('Relative Uncertainties%s', plotTitleSuffix));
        grid on;

    catch ME_plot
        warning('Could not create confidence interval plots: %s', ME_plot.message);
    end
else
     warning('Skipping CI plots because CI calculation failed or fit did not converge.');
end

% T1 plot - scaled to exp data
figure("Color", "White")
loglog(B, [1./R1_J_fit, 1./R1_DD_fit, 1./R1_CSA_fit, 1./R1_fit_total], "LineWidth", 3);
xlabel("Magnetic field, T")
ylabel("T_1 (s)")
hold on
loglog(B_exp, T1_exp, "o", "MarkerSize", 10, "LineWidth", 2)
hold off
title(sprintf('T_1 (scaled to exp. data%s) - R² = %.4f', plotTitleSuffix, R2))
legend(["T_{1J}", "T_{1DD}", "T_{1CSA}", "T_{1,all}", "Exp data"])
grid on; xlim([min(B_exp)*0.9, max(B_exp)*1.1]); ylim([min(T1_exp)*0.9, max(T1_exp)*1.1]);

%% Helper Functions

% *** Modified DYNAMIC Helper Function for UNSCALED parameters ***
function R = calculate_R1_dynamic_unscaled(params_unscaled_fit, params_base_unscaled, fixedFlags, constrainTauDtoTauCSA, B, gamma15N, gamma1H) % Added flag
    % Reconstruct the full UNSCALED parameter vector
    params_all_unscaled = params_base_unscaled;
    params_all_unscaled(~fixedFlags) = params_unscaled_fit;

    % Extract unscaled parameters by index
    J       = params_all_unscaled(1);
    tauJ    = params_all_unscaled(2);
    tauD    = params_all_unscaled(3); % Represents tauC if constrained
    tauCSA  = params_all_unscaled(4); % Base value (fixed if constrained)
    CSA     = params_all_unscaled(5);
    rIS     = params_all_unscaled(6);
    T10     = params_all_unscaled(7);

    % --- Apply tauD = tauCSA constraint if requested ---
    if constrainTauDtoTauCSA
        tauCSA = tauD; % Force tauCSA to equal the fitted tauD
    end
    % ----------------------------------------------------

    % Basic checks
    if T10 <= 0 || tauJ <= 0 || tauD <= 0 || tauCSA <= 0 || J < 0 || CSA < 0
        R = ones(size(B)) * 1e10;
        return;
    end
    R10 = 1/T10;
    S = 1/2;

    % Calculate R1 components safely
    try
        R1_J_val = R_J(J, S, gamma15N, gamma1H, tauJ, B);
        R1_DD_val = R_DD(rIS, S, gamma15N, gamma1H, tauD, B); % Uses fitted tauD (tauC)
        R1_CSA_val = R_CSA(gamma15N*B, CSA, tauCSA); % Uses tauCSA (which equals tauD if constrained)
        R_components = R10 + R1_J_val + R1_DD_val + R1_CSA_val;

        if any(isnan(R_components)) || any(isinf(R_components)) || any(R_components < 0)
            warning('calculate_R1_dynamic: Invalid R1 value calculated. Penalizing.');
            R = ones(size(B)) * 1e10;
        else
            R = R_components;
        end
    catch ME_calc
         warning('calculate_R1_dynamic: Error during R1 calculation: %s. Penalizing.', ME_calc.message);
         R = ones(size(B)) * 1e10;
    end
end


% ... (R_J, R_DD, R_CSA, Jw0 functions remain the same) ...
function R = R_J(J, S, gammaI, gammaS, tau, B)
    if isscalar(tau); tau = repmat(tau, size(B)); end
    if isscalar(B); B = repmat(B, size(tau)); end
    R = 8*pi^2*J^2/3*(S*(S+1))*tau.*Jw0((gammaI-gammaS)*B.*tau);
end

function R = R_CSA(w0, CSA, tau)
    if isscalar(tau); tau = repmat(tau, size(w0)); end
    if isscalar(w0); w0 = repmat(w0, size(tau)); end
    R = (1/5)*(w0*CSA).^2.*tau.*Jw0(w0.*tau);
end

function R = R_DD(rIS, S, gammaI, gammaS, tau, B)
    mu0 = 1.25663706127e-6; % N*A-2
    hbar = 1.054571817e-34; % J⋅s
    if isscalar(tau); tau = repmat(tau, size(B)); end
    if isscalar(B); B = repmat(B, size(tau)); end
    R = 2/3*(mu0*hbar*gammaI*gammaS/(4*pi*rIS^3))^2*S*(S+1)*tau.*(Jw0((gammaI-gammaS)*B.*tau) + 3*Jw0(gammaI*B.*tau) + 6*Jw0((gammaI+gammaS)*B.*tau));
end

function J = Jw0(x)
    J = 1./(1+(x).^2);
end
