% [1] Chiavazza et al 10.1016/j.jmr.2012.11.016
% [2] Kharkov et al 10.1039/d1cp05537b

clear all; close all; clc; % Start fresh

%% Parameter Configuration =============================================
% --- Constraint Flags ---
constrainTauJ = false; % SET TO true TO FORCE tauJ1 = tauJ2

% --- Define ALL potential parameters and their initial guesses ---
paramList = {'J1', 'J2', 'tauJ1', 'tauJ2', 'tauD', 'tauCSA', 'CSA', 'rIS', 'T10'};
initialGuesses = [
    45,     % J1 (Hz)
    80,     % J2 (Hz)
    5e-7,   % tauJ1 (s)
    3e-9,   % tauJ2 (s) - Initial guess still needed even if constrained
    8e-12,  % tauD (s)
    6.6e-11, % tauCSA (s)
    1.5e-4,  % CSA (unitless)
    2.3e-10,% rIS (m)
    28      % T10 (s) - Adjusted guess based on data
];

% --- Define which parameters to FIX and their FIXED VALUES ---
fixedFlags = [
    false,  % J1
    false,  % J2
    false,  % tauJ1
    false,  % tauJ2 - Will be overridden if constrainTauJ is true
    true,   % tauD
    false,  % tauCSA
    false,  % CSA
    true,   % rIS
    false   % T10
];
fixedValues = initialGuesses;
fixedValues(strcmp(paramList, 'J1')) = 0;
fixedValues(strcmp(paramList, 'J2')) = 0;
fixedValues(strcmp(paramList, 'tauJ1')) = 1e-30;
fixedValues(strcmp(paramList, 'tauJ2')) = 1e-30;
fixedValues(strcmp(paramList, 'rIS')) = 1e-10;
fixedValues(strcmp(paramList, 'tauD')) = 1e-30;



% --- Apply tauJ constraint if requested ---
if constrainTauJ
    tauJ2_idx = find(strcmp(paramList, 'tauJ2'));
    if ~isempty(tauJ2_idx)
        fixedFlags(tauJ2_idx) = true; % Fix tauJ2 if constraining
        fprintf('Constraint Applied: tauJ2 will be forced equal to tauJ1.\n');
    end
end

% --- Define Bounds for ALL parameters ---
lowerBounds = [1,    1,    1e-9,  1e-10, 1e-14,  1e-14,  10e-6,  1.8e-10, 1];
upperBounds = [1000, 1000, 1e-5,  1e-6,  1e-10,  1e-10,  500e-6, 2.8e-10, 200];

% --- REMOVED SCALING FACTORS ---
% =====================================================================

%% Constants
S_init = 1/2;
gamma1H = 26.75222080e7;
gamma15N = -2.7126189e7;

%% Experimental data
D = [
0.0000625	17.62115
0.00009375	17.86318
0.000125	17.7492
0.0001875	17.8994
0.00025	18.37386
0.000375	17.75006
0.0005	17.56465
0.00075	18.30326
0.001	17.66549
0.0015	17.61455
0.002	17.52801
0.003	19.29173
0.004	20.70913
0.006	22.05391
0.008	22.3085
0.012	25.56833
0.016	25.03465
0.024	26.47779
0.032	26.74396
0.048	27.96799
0.064	27.13301
0.096	28.16546
0.128	27.94884
0.192	26.9685
0.256	27.42236
0.384	27.85787
0.512	28.13478
0.768	27.77281
1.024	28.43833
1.536	26.9251
2.048	27.53716
3.072	27.16383
4.096	25.17265
6.144	22.14
8.192	19.56489
9.4	17.45499

];

B_exp = D(:,1);
T1_exp = D(:,2);
R1_exp = 1./T1_exp;
B = logspace(log10(min(B_exp)*0.1), log10(max(B_exp)*1.1), 1000)';

%% Prepare for Fitting (NO SCALING)
fitIndices = find(~fixedFlags);
fixedIndices = find(fixedFlags);

params_base_unscaled = fixedValues;
% If tauJ2 is constrained, make sure its value in the base vector is consistent (though it will be overwritten)
if constrainTauJ
    tauJ1_idx = find(strcmp(paramList, 'tauJ1'));
    tauJ2_idx = find(strcmp(paramList, 'tauJ2'));
    if ~isempty(tauJ1_idx) && ~isempty(tauJ2_idx)
        params_base_unscaled(tauJ2_idx) = params_base_unscaled(tauJ1_idx);
    end
end
params_base_unscaled(fitIndices) = initialGuesses(fitIndices); % Use initial guesses for non-fixed, non-constrained params

params0_fit_unscaled = params_base_unscaled(fitIndices);
lb_fit_unscaled = lowerBounds(fitIndices);
ub_fit_unscaled = upperBounds(fitIndices);

%% Perform the fitting (NO SCALING)
options = optimoptions('lsqnonlin', ...
    'Display', 'iter', ...
    'MaxFunctionEvaluations', 5000, ...
    'MaxIterations', 1000, ...
    'FunctionTolerance', 1e-9, ...
    'StepTolerance', 1e-9, ...
    'OptimalityTolerance', 1e-9, ...
    'Algorithm', 'trust-region-reflective', ...
    'FiniteDifferenceType', 'central', ...
    'UseParallel', false);

% Pass constrainTauJ flag to the objective function
objFun_unscaled = @(p_unscaled_fit) calculate_R1_dynamic_unscaled(p_unscaled_fit, params_base_unscaled, fixedFlags, constrainTauJ, B_exp, gamma15N, gamma1H) - R1_exp;

[params_fit_unscaled, resnorm, residual, exitflag, output, ~, jacobian_unscaled] = lsqnonlin(objFun_unscaled, params0_fit_unscaled, lb_fit_unscaled, ub_fit_unscaled, options);

% Condition number check
if exist('jacobian_unscaled','var') && ~isempty(jacobian_unscaled)
    try
        jacobian_full = full(jacobian_unscaled);
        cond_num = cond(jacobian_full' * jacobian_full);
        fprintf('Condition number of J''*J: %.2e\n', cond_num);
    catch cond_err
        fprintf('Could not calculate condition number: %s\n', cond_err.message);
    end
else
    fprintf('Jacobian not available for condition number check.\n');
end


%% Process Results (NO SCALING)
params_all_final = params_base_unscaled;
params_all_final(fitIndices) = params_fit_unscaled;
% If constrained, ensure final tauJ2 equals final tauJ1
if constrainTauJ
    tauJ1_idx = find(strcmp(paramList, 'tauJ1'));
    tauJ2_idx = find(strcmp(paramList, 'tauJ2'));
     if ~isempty(tauJ1_idx) && ~isempty(tauJ2_idx)
        params_all_final(tauJ2_idx) = params_all_final(tauJ1_idx);
     end
end

finalValues = struct();
for i = 1:length(paramList)
    finalValues.(paramList{i}) = params_all_final(i);
end

% Confidence intervals
param_errors_fit = nan(size(params_fit_unscaled));
ci_calculation_ok = false;
if exitflag > 0
    try
        ci_unscaled_fit = nlparci(params_fit_unscaled, residual, 'jacobian', jacobian_unscaled);
        param_errors_fit = (ci_unscaled_fit(:,2) - ci_unscaled_fit(:,1)) / 2;
        ci_calculation_ok = true;
    catch ME
        warning('Could not calculate confidence intervals using nlparci: %s', ME.message);
        if exist('jacobian_unscaled','var') && ~isempty(jacobian_unscaled)
             try
                 cond_num = cond(full(jacobian_unscaled)' * full(jacobian_unscaled));
                 fprintf('Condition number of J''*J: %.2e\n', cond_num);
             catch cond_err
                 fprintf('Could not calculate condition number: %s\n', cond_err.message);
             end
        end
    end
end

% Display results
fprintf('\n--- Fit Results ---\n');
R2 = 1 - resnorm/sum((R1_exp - mean(R1_exp)).^2);
fprintf('R-squared: %.4f\n', R2);
fprintf('Exit Flag: %d, Message: %s\n', exitflag, output.message);
if constrainTauJ; fprintf('Constraint Applied: tauJ1 = tauJ2\n'); end
fprintf('\n--- Final Parameter Values and 95%% CIs ---\n');
err_idx = 1;
for i = 1:length(paramList)
    pName = paramList{i}; pValue = finalValues.(pName);
    % Special handling for constrained tauJ2 display
    isTauJ2Constrained = constrainTauJ && strcmp(pName, 'tauJ2');

    if fixedFlags(i) && ~isTauJ2Constrained % Fixed parameters (excluding constrained tauJ2)
        fprintf('%s (fixed) = %.3e\n', pName, pValue);
    elseif isTauJ2Constrained % Constrained tauJ2
        fprintf('%s (constrained to tauJ1) = %.3e\n', pName, pValue);
    else % Fitted parameters
        if ci_calculation_ok && err_idx <= length(param_errors_fit)
            pError = param_errors_fit(err_idx);
            fprintf('%s = %.3e ± %.3e\n', pName, pValue, pError);
        else
            fprintf('%s = %.3e ± (CI failed)\n', pName, pValue);
        end
        err_idx = err_idx + 1;
    end
end

% Check bounds
fprintf('\nChecking if fitted parameters are close to bounds:\n');
tol = 1e-5; fit_idx_counter = 1;
for i = 1:length(paramList)
    % Only check parameters that were actually fitted by lsqnonlin
    if ~fixedFlags(i)
        param_name = paramList{i}; fitted_value = finalValues.(param_name);
        lb_val = lowerBounds(i); ub_val = upperBounds(i);
        if fitted_value <= lb_val * (1 + tol) + lb_val*tol
             fprintf('Parameter %s (%.3e) is close to lower bound (%.3e)\n', param_name, fitted_value, lb_val);
        elseif fitted_value >= ub_val * (1 - tol) - ub_val*tol
             fprintf('Parameter %s (%.3e) is close to upper bound (%.3e)\n', param_name, fitted_value, ub_val);
        end
        fit_idx_counter = fit_idx_counter + 1;
    end
end

%% Calculate relaxation rates & Contributions
R1_J1_fit = R_J(finalValues.J1, S_init, gamma15N, gamma1H, finalValues.tauJ1, B);
R1_J2_fit = R_J(finalValues.J2, S_init, gamma15N, gamma1H, finalValues.tauJ2, B);
R1_J_fit_total = R1_J1_fit + R1_J2_fit; % Still useful for overall R1

R1_DD_fit = R_DD(finalValues.rIS, S_init, gamma15N, gamma1H, finalValues.tauD, B);
R1_CSA_fit = R_CSA(gamma15N*B, finalValues.CSA, finalValues.tauCSA);
R10_fit = 1/finalValues.T10;
R1_fit_total = R10_fit + R1_J_fit_total + R1_DD_fit + R1_CSA_fit;

% Calculate R1 at experimental points
R1_J1_exp = R_J(finalValues.J1, S_init, gamma15N, gamma1H, finalValues.tauJ1, B_exp);
R1_J2_exp = R_J(finalValues.J2, S_init, gamma15N, gamma1H, finalValues.tauJ2, B_exp);
R1_DD_exp = R_DD(finalValues.rIS, S_init, gamma15N, gamma1H, finalValues.tauD, B_exp);
R1_CSA_exp = R_CSA(gamma15N*B_exp, finalValues.CSA, finalValues.tauCSA);
R1_fit_at_exp = R10_fit + R1_J1_exp + R1_J2_exp + R1_DD_exp + R1_CSA_exp;

% Contributions
fprintf('\nRelative contributions at 1T:\n');
R1_J1_1T = R_J(finalValues.J1, S_init, gamma15N, gamma1H, finalValues.tauJ1, 1);
R1_J2_1T = R_J(finalValues.J2, S_init, gamma15N, gamma1H, finalValues.tauJ2, 1);
R1_J_total_1T = R1_J1_1T + R1_J2_1T;
R1_DD_1T = R_DD(finalValues.rIS, S_init, gamma15N, gamma1H, finalValues.tauD, 1);
R1_CSA_1T = R_CSA(gamma15N*1, finalValues.CSA, finalValues.tauCSA);
R1_total_1T = R10_fit + R1_J_total_1T + R1_DD_1T + R1_CSA_1T;
fprintf('R10: %.4e s^-1 (%.1f%%)\n', R10_fit, 100*R10_fit/R1_total_1T);
fprintf('R1_J1: %.4e s^-1 (%.1f%%)\n', R1_J1_1T, 100*R1_J1_1T/R1_total_1T); % Individual J1
fprintf('R1_J2: %.4e s^-1 (%.1f%%)\n', R1_J2_1T, 100*R1_J2_1T/R1_total_1T); % Individual J2
fprintf('R1_DD: %.4e s^-1 (%.1f%%)\n', R1_DD_1T, 100*R1_DD_1T/R1_total_1T);
fprintf('R1_CSA: %.4e s^-1 (%.1f%%)\n', R1_CSA_1T, 100*R1_CSA_1T/R1_total_1T);
fprintf('R1_total: %.4e s^-1\n', R1_total_1T);

fprintf('\nRelative contributions at 9.4T:\n');
R1_J1_9T = R_J(finalValues.J1, S_init, gamma15N, gamma1H, finalValues.tauJ1, 9.4);
R1_J2_9T = R_J(finalValues.J2, S_init, gamma15N, gamma1H, finalValues.tauJ2, 9.4);
R1_J_total_9T = R1_J1_9T + R1_J2_9T;
R1_DD_9T = R_DD(finalValues.rIS, S_init, gamma15N, gamma1H, finalValues.tauD, 9.4);
R1_CSA_9T = R_CSA(gamma15N*9.4, finalValues.CSA, finalValues.tauCSA);
R1_total_9T = R10_fit + R1_J_total_9T + R1_DD_9T + R1_CSA_9T;
fprintf('R10: %.4e s^-1 (%.1f%%)\n', R10_fit, 100*R10_fit/R1_total_9T);
fprintf('R1_J1: %.4e s^-1 (%.1f%%)\n', R1_J1_9T, 100*R1_J1_9T/R1_total_9T); % Individual J1
fprintf('R1_J2: %.4e s^-1 (%.1f%%)\n', R1_J2_9T, 100*R1_J2_9T/R1_total_9T); % Individual J2
fprintf('R1_DD: %.4e s^-1 (%.1f%%)\n', R1_DD_9T, 100*R1_DD_9T/R1_total_9T);
fprintf('R1_CSA: %.4e s^-1 (%.1f%%)\n', R1_CSA_9T, 100*R1_CSA_9T/R1_total_9T);
fprintf('R1_total: %.4e s^-1\n', R1_total_9T);

%% Plotting
close all
fixedParamStrParts = {};
constrainedParamStrParts = {};
for i = 1:length(paramList)
    isTauJ2Constrained = constrainTauJ && strcmp(paramList{i}, 'tauJ2');
    if fixedFlags(i) && ~isTauJ2Constrained
        fixedParamStrParts{end+1} = paramList{i};
    elseif isTauJ2Constrained
        constrainedParamStrParts{end+1} = 'tauJ1=tauJ2';
    end
end
fixedParamStr = strjoin(fixedParamStrParts, ', ');
constrainedParamStr = strjoin(constrainedParamStrParts, ', ');
plotTitleSuffix = '';
if ~isempty(fixedParamStr); plotTitleSuffix = [plotTitleSuffix, ', Fixed: ', fixedParamStr]; end
if ~isempty(constrainedParamStr); plotTitleSuffix = [plotTitleSuffix, ', Constrained: ', constrainedParamStr]; end


% R1 plot - Show individual J components
figure("Color", "White")
loglog(B, [R1_J1_fit, R1_J2_fit, R1_DD_fit, R1_CSA_fit, R1_fit_total], "LineWidth", 3); % Plot J1, J2 separately
xlabel("Magnetic field, T")
ylabel("R_1 (s^{-1})")
hold on
loglog(B_exp, R1_exp, "o", "MarkerSize", 10, "LineWidth", 2)
hold off
title(sprintf('R_1 (all contributions%s)', plotTitleSuffix))
legend(["R_{1,J1}", "R_{1,J2}", "R_{1DD}", "R_{1CSA}", "R_{1,all}", "Exp data"]) % Updated legend
grid on; xlim([min(B_exp)*0.9, max(B_exp)*1.1]);

% T1 plot - Show individual J components
figure("Color", "White")
T1_J1_fit = 1./R1_J1_fit; T1_J1_fit(isinf(T1_J1_fit)) = NaN;
T1_J2_fit = 1./R1_J2_fit; T1_J2_fit(isinf(T1_J2_fit)) = NaN;
loglog(B, [T1_J1_fit, T1_J2_fit, 1./R1_DD_fit, 1./R1_CSA_fit, 1./R1_fit_total], "LineWidth", 3); % Plot T1J1, T1J2
xlabel("Magnetic field, T")
ylabel("T_1 (s)")
hold on
loglog(B_exp, T1_exp, "o", "MarkerSize", 10, "LineWidth", 2)
hold off
title(sprintf('T_1 (all contributions%s)', plotTitleSuffix))
legend(["T_{1,J1}", "T_{1,J2}", "T_{1DD}", "T_{1CSA}", "T_{1,all}", "Exp data"]) % Updated legend
grid on; xlim([min(B_exp)*0.9, max(B_exp)*1.1]);

% R1 plot - scaled to exp data
figure("Color", "White")
loglog(B, [R1_J1_fit, R1_J2_fit, R1_DD_fit, R1_CSA_fit, R1_fit_total], "LineWidth", 3); % Plot J1, J2
xlabel("Magnetic field, T")
ylabel("R_1 (s^{-1})")
hold on
loglog(B_exp, R1_exp, "o", "MarkerSize", 10, "LineWidth", 2)
hold off
title(sprintf('R_1 (scaled to exp. data%s) - R² = %.4f', plotTitleSuffix, R2))
legend(["R_{1,J1}", "R_{1,J2}", "R_{1DD}", "R_{1CSA}", "R_{1,all}", "Exp data"]) % Updated legend
grid on; xlim([min(B_exp)*0.9, max(B_exp)*1.1]); ylim([min(R1_exp)*0.9, max(R1_exp)*1.1]);

% Residuals plot
figure("Color", "White")
semilogx(B_exp, R1_exp - R1_fit_at_exp, 'o', 'MarkerSize', 10, 'LineWidth', 2);
xlabel("Magnetic field, T")
ylabel("Residuals (s^{-1})")
title(sprintf('Fit Residuals%s', plotTitleSuffix))
grid on
line([min(B_exp)*0.9, max(B_exp)*1.1], [0, 0], 'Color', 'k', 'LineStyle', '--')

% Confidence Interval Plots
if ci_calculation_ok
    try
        fittedParamNames = paramList(fitIndices);
        fittedParamValues = params_all_final(fitIndices);

        figure("Color", "White")
        errorbar(1:length(fittedParamValues), fittedParamValues, param_errors_fit, 'o', 'MarkerSize', 10, 'LineWidth', 2, 'LineStyle', 'none');
        set(gca, 'XTick', 1:length(fittedParamValues), 'XTickLabel', fittedParamNames);
        ylabel('Parameter Value');
        title(sprintf('Fitted Parameters CIs%s', plotTitleSuffix));
        grid on;
        set(gca,'YScale','log')

        figure("Color", "White")
        relative_uncertainty_fit = param_errors_fit ./ abs(fittedParamValues);
        bar(1:length(fittedParamValues), relative_uncertainty_fit);
        set(gca, 'XTick', 1:length(fittedParamValues), 'XTickLabel', fittedParamNames);
        ylabel('Relative Uncertainty (CI half-width / value)');
        title(sprintf('Relative Uncertainties%s', plotTitleSuffix));
        grid on;

    catch ME_plot
        warning('Could not create confidence interval plots: %s', ME_plot.message);
    end
else
     warning('Skipping CI plots because CI calculation failed or fit did not converge.');
end

% T1 plot - scaled to exp data
figure("Color", "White")
loglog(B, [T1_J1_fit, T1_J2_fit, 1./R1_DD_fit, 1./R1_CSA_fit, 1./R1_fit_total], "LineWidth", 3); % Plot T1J1, T1J2
xlabel("Magnetic field, T")
ylabel("T_1 (s)")
hold on
loglog(B_exp, T1_exp, "o", "MarkerSize", 10, "LineWidth", 2)
hold off
title(sprintf('T_1 (scaled to exp. data%s) - R² = %.4f', plotTitleSuffix, R2))
legend(["T_{1,J1}", "T_{1,J2}", "T_{1DD}", "T_{1CSA}", "T_{1,all}", "Exp data"]) % Updated legend
grid on; xlim([min(B_exp)*0.9, max(B_exp)*1.1]); ylim([min(T1_exp)*0.9, max(T1_exp)*1.1]);


%% Helper Functions

% *** Modified DYNAMIC Helper Function for UNSCALED parameters ***
function R = calculate_R1_dynamic_unscaled(params_unscaled_fit, params_base_unscaled, fixedFlags, constrainTauJ, B, gamma15N, gamma1H)
    % Reconstruct the full UNSCALED parameter vector
    params_all_unscaled = params_base_unscaled;
    params_all_unscaled(~fixedFlags) = params_unscaled_fit;

    % Extract unscaled parameters by index
    J1      = params_all_unscaled(1);
    J2      = params_all_unscaled(2);
    tauJ1   = params_all_unscaled(3);
    tauJ2   = params_all_unscaled(4);
    tauD    = params_all_unscaled(5);
    tauCSA  = params_all_unscaled(6);
    CSA     = params_all_unscaled(7);
    rIS     = params_all_unscaled(8);
    T10     = params_all_unscaled(9);

    % --- Apply tauJ constraint if requested ---
    if constrainTauJ
        tauJ2 = tauJ1;
    end
    % -----------------------------------------

    % Basic checks
    if T10 <= 0 || tauJ1 <= 0 || tauJ2 <= 0 || tauD <= 0 || tauCSA <= 0
        R = ones(size(B)) * 1e10;
        return;
    end
    R10 = 1/T10;
    S = 1/2;

    % Calculate R1 components safely
    try
        R1_J1_val = R_J(J1, S, gamma15N, gamma1H, tauJ1, B);
        R1_J2_val = R_J(J2, S, gamma15N, gamma1H, tauJ2, B); % Uses potentially constrained tauJ2
        R1_DD_val = R_DD(rIS, S, gamma15N, gamma1H, tauD, B);
        R1_CSA_val = R_CSA(gamma15N*B, CSA, tauCSA);
        R_components = R10 + R1_J1_val + R1_J2_val + R1_DD_val + R1_CSA_val;

        if any(isnan(R_components)) || any(isinf(R_components)) || any(R_components < 0)
            warning('calculate_R1_dynamic: Invalid R1 value calculated. Penalizing.');
            R = ones(size(B)) * 1e10;
        else
            R = R_components;
        end
    catch ME_calc
         warning('calculate_R1_dynamic: Error during R1 calculation: %s. Penalizing.', ME_calc.message);
         R = ones(size(B)) * 1e10;
    end
end

% --- Original helper functions R_J, R_CSA, R_DD, Jw0 (no changes needed) ---
% ... (Include R_J, R_CSA, R_DD, Jw0 functions here) ...
function R = R_J(J, S, gammaI, gammaS, tau, B)
    if isscalar(tau); tau = repmat(tau, size(B)); end
    if isscalar(B); B = repmat(B, size(tau)); end
    R = 8*pi^2*J^2/3*(S*(S+1))*tau.*Jw0((gammaI-gammaS)*B.*tau);
end

function R = R_CSA(w0, CSA, tau)
    if isscalar(tau); tau = repmat(tau, size(w0)); end
    if isscalar(w0); w0 = repmat(w0, size(tau)); end
    R = (1/5)*(w0*CSA).^2.*tau.*Jw0(w0.*tau);
end

function R = R_DD(rIS, S, gammaI, gammaS, tau, B)
    mu0 = 1.25663706127e-6; % N*A-2
    hbar = 1.054571817e-34; % J⋅s
    if isscalar(tau); tau = repmat(tau, size(B)); end
    if isscalar(B); B = repmat(B, size(tau)); end
    R = 2/3*(mu0*hbar*gammaI*gammaS/(4*pi*rIS^3))^2*S*(S+1)*tau.*(Jw0((gammaI-gammaS)*B.*tau) + 3*Jw0(gammaI*B.*tau) + 6*Jw0((gammaI+gammaS)*B.*tau));
end

function J = Jw0(x)
    J = 1./(1+(x).^2);
end